/**
 * Dream Copyright (C) 2003-2004 INRIA Rhone-Alpes
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option) any
 * later version.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * Contact: dream@objectweb.org
 * 
 * Initial developer(s): Matthieu Leclercq 
 * Contributor(s):
 */

package org.objectweb.dream.control.activity.task.thread;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;

import org.objectweb.dream.control.activity.task.Task;
import org.objectweb.dream.control.activity.task.TaskLifeCycleController;
import org.objectweb.dream.control.activity.task.TaskStoppedListener;
import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.control.IllegalLifeCycleException;
import org.objectweb.fractal.julia.control.lifecycle.ChainedIllegalLifeCycleException;
import org.objectweb.util.monolog.api.BasicLevel;

/**
 * Thread Pool implementation. In a thread pool, many threads call concurrently
 * the scheduler. The number of threads can be controlled by the
 * {@link ThreadPoolController }control interface.
 */
public class ThreadPoolTask extends AbstractThreadTask
    implements
      TaskLifeCycleController,
      ThreadPoolController
{

  Object     threadTaskLock       = new Object();
  long       waitTimeout          = DEFAULT_WAIT_TIMEOUT;
  int        capacity             = 0;
  LinkedList activeThreads        = new LinkedList();
  Set        waitingThreads       = new HashSet();

  List       taskStoppedListeners = new ArrayList();
  boolean    stopping;

  /**
   * @see AbstractThreadTask#isExecuting()
   */
  protected boolean isExecuting()
  {
    return ((PoolThread) Thread.currentThread()).executing;
  }

  /**
   * @see AbstractThreadTask#setExecuting(boolean)
   */
  protected void setExecuting(boolean b)
  {
    ((PoolThread) Thread.currentThread()).executing = b;
  }

  // ---------------------------------------------------------------------------
  // Implementation of the ThreadPoolController interface
  // ---------------------------------------------------------------------------

  /**
   * @see ThreadPoolController#getNbActiveThreads()
   */
  public int getNbActiveThreads()
  {
    synchronized (threadTaskLock)
    {
      return activeThreads.size();
    }
  }

  /**
   * @see ThreadPoolController#addThreads(int)
   */
  public void addThreads(int i) throws ThreadPoolOverflowException,
      IllegalLifeCycleException
  {
    synchronized (threadTaskLock)
    {
      if (getFcState() == STOPPED)
      {
        throw new IllegalLifeCycleException(
            "Can't add thread in the stopped state.");
      }
      if (getNbActiveThreads() + i > getCapacity())
      {
        throw new ThreadPoolOverflowException(
            "Can't add Thread in pool due to capacity limitation");
      }
      for (int j = 0; j < i; j++)
      {
        PoolThread t = null;
        synchronized (waitingThreads)
        {
          if (!waitingThreads.isEmpty())
          {
            // notify a waiting thread
            Iterator iterator = waitingThreads.iterator();
            t = (PoolThread) iterator.next();
            iterator.remove();
            synchronized (activeThreads)
            {
              activeThreads.add(t);
            }
            synchronized (t)
            {
              logger.log(BasicLevel.DEBUG, "Notify a waiting thread task");
              t.revive = true;
              t.notify();
            }
            continue;
          }
        }
        logger.log(BasicLevel.DEBUG, "Creates a new thread task");
        t = new PoolThread();
        synchronized (activeThreads)
        {
          activeThreads.add(t);
        }
        t.start();
      }
    }
  }

  /**
   * @see ThreadPoolController#removeThreads(int)
   */
  public void removeThreads(int i) throws IllegalLifeCycleException
  {
    synchronized (threadTaskLock)
    {
      if (getFcState() == STOPPED)
      {
        throw new IllegalLifeCycleException(
            "Can't remove thread in the stopped state.");
      }
      for (int j = 0; j < i; j++)
      {
        PoolThread t;
        synchronized (activeThreads)
        {
          t = (PoolThread) activeThreads.removeFirst();
        }
        logger.log(BasicLevel.DEBUG, "Stop a thread task");
        t.executing = false;
        t.interrupt();
      }
    }
  }

  /**
   * @see ThreadPoolController#setCapacity(int)
   */
  public void setCapacity(int i)
  {
    capacity = i;
  }

  /**
   * @see ThreadPoolController#getCapacity()
   */
  public int getCapacity()
  {
    return capacity;
  }

  /**
   * @see ThreadPoolController#getWaitTimeout()
   */
  public long getWaitTimeout()
  {
    return waitTimeout;
  }

  /**
   * @see ThreadPoolController#setWaitTimeout(long)
   */
  public void setWaitTimeout(long millis)
  {
    waitTimeout = millis;
  }

  // ---------------------------------------------------------------------------
  // Implementation of the LifeCycleController interface
  // ---------------------------------------------------------------------------

  /**
   * @see org.objectweb.fractal.api.control.LifeCycleController#startFc()
   */
  public void startFc() throws IllegalLifeCycleException
  {
    if (getFcState() == STARTED)
    {
      return;
    }
    synchronized (threadTaskLock)
    {
      super.startFc();
      stopping = false;
    }
  }

  /**
   * @see org.objectweb.fractal.api.control.LifeCycleController#stopFc()
   */
  public void stopFc() throws IllegalLifeCycleException
  {
    if (getFcState() == STOPPED)
    {
      return;
    }
    synchronized (threadTaskLock)
    {
      logger.log(BasicLevel.DEBUG, "Stopping thread pool task");
      stopping = true;
      interruptPool();
      boolean noActiveThread;
      synchronized (activeThreads)
      {
        noActiveThread = activeThreads.isEmpty();
      }
      while (!noActiveThread)
      {
        try
        {
          logger.log(BasicLevel.DEBUG, "Join threads of the task");
          threadTaskLock.wait();
        }
        catch (InterruptedException e)
        {
          throw new ChainedIllegalLifeCycleException(e, weaveableC,
              "Interrupted while waiting for the end of the thread pool.");
        }
        synchronized (activeThreads)
        {
          noActiveThread = activeThreads.isEmpty();
        }
      }
    }
  }

  /**
   * @see TaskLifeCycleController#asyncStop(TaskStoppedListener)
   */
  public void asyncStop(TaskStoppedListener listener)
  {
    if (getFcState() == STOPPED)
    {
      //    thread is already stopped
      Task taskItf = null;
      try
      {
        taskItf = (Task) weaveableC.getFcInterface("task");
      }
      catch (NoSuchInterfaceException ignored)
      {
        // can't happend
      }
      listener.taskStopped(taskItf);
      return;
    }
    synchronized (threadTaskLock)
    {
      logger.log(BasicLevel.DEBUG, "Stopping asynchronously thread pool task");
      stopping = true;
      interruptPool();
      taskStoppedListeners.add(listener);
    }
  }

  // ---------------------------------------------------------------------------
  // Utility method
  // ---------------------------------------------------------------------------

  protected void interruptPool()
  {
    logger.log(BasicLevel.DEBUG, "Interrupting threads of the task");
    synchronized (activeThreads)
    {
      Iterator iter = activeThreads.iterator();
      while (iter.hasNext())
      {
        PoolThread thread = (PoolThread) iter.next();
        thread.executing = false;
        thread.interrupt();
      }
    }
    synchronized (waitingThreads)
    {
      Iterator iter = waitingThreads.iterator();
      while (iter.hasNext())
      {
        PoolThread thread = (PoolThread) iter.next();
        thread.executing = false;
        thread.interrupt();
      }
    }
  }

  // ---------------------------------------------------------------------------
  // Inner class
  // ---------------------------------------------------------------------------

  private class PoolThread extends Thread
  {

    boolean executing = true;
    boolean revive;

    // ---------------------------------------------------------------------------
    // Implementation of the Runnable interface
    // ---------------------------------------------------------------------------

    /**
     * @see Runnable#run()
     */
    public void run()
    {
      logger.log(BasicLevel.DEBUG, "Begin of the run method");
      boolean runThread = true;
      do
      {
        revive = false;
        try
        {
          execute(null);
        }
        catch (InterruptedException e)
        {
          logger.log(BasicLevel.DEBUG, "Interrupted");
        }
        finally
        {
          boolean noActiveThread;
          synchronized (activeThreads)
          {
            // remove myself from activeThreads.
            activeThreads.remove(this);
            if (activeThreads.size() >= getCapacity())
            {
              logger.log(BasicLevel.DEBUG, "More active threads than capacity");
              runThread = false;
            }
            noActiveThread = activeThreads.isEmpty();
          }
          List listeners = null;
          synchronized (threadTaskLock)
          {
            if (stopping && noActiveThread)
            {
              logger.log(BasicLevel.DEBUG,
                  "No more active thread component stopped");
              try
              {
                // set state to stop.
                ThreadPoolTask.super.stopFc();
                stopping = false;
              }
              catch (IllegalLifeCycleException ignored)
              {
                // ignored
              }
              // notify threads that may be waiting in stopFc.
              threadTaskLock.notifyAll();
              listeners = taskStoppedListeners;
              taskStoppedListeners = new ArrayList();
            }
          }
          if (listeners != null)
          {
            // notify listener.
            Task taskItf = null;
            try
            {
              taskItf = (Task) weaveableC.getFcInterface("task");
            }
            catch (NoSuchInterfaceException ignored)
            {
              // can't happend
            }
            Iterator iter = listeners.iterator();
            while (iter.hasNext())
            {
              TaskStoppedListener listener = (TaskStoppedListener) iter.next();
              listener.taskStopped(taskItf);
            }
            // end run method.
            break;
          }
          synchronized (waitingThreads)
          {
            // add myself in waitingThreads.
            waitingThreads.add(this);
          }
          try
          {
            synchronized (this)
            {
              if (!revive && runThread)
              {
                try
                {
                  logger.log(BasicLevel.DEBUG, "Waits for revive");
                  this.wait(waitTimeout);
                  executing = true;
                }
                catch (InterruptedException e1)
                {
                  logger.log(BasicLevel.DEBUG,
                      "Interrupted while waiting for timeout");
                }
              }
            }
          }
          finally
          {
            synchronized (waitingThreads)
            {
              // remove myself in waitingThreads.
              waitingThreads.remove(this);
            }
          }
        }
      }
      while (revive && runThread);
      logger.log(BasicLevel.DEBUG, "End of the run method");
    }
  }
}