/**
 * Dream
 * Copyright (C) 2003-2004 INRIA Rhone-Alpes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact : dream@objectweb.org
 *
 * Initial developer(s): Matthieu Leclercq
 * Contributor(s): Vivien Quema
 */

package org.objectweb.dream.message;

import java.io.Externalizable;
import java.io.IOException;
import java.io.ObjectInput;
import java.io.ObjectOutput;

import org.objectweb.dream.pool.Recyclable;

/**
 * Basic implementation of {@link org.objectweb.dream.message.ChunkType}
 * interface
 */
public class ChunkTypeImpl implements ChunkType, Recyclable, Externalizable
{

  private Class chunkItf;
  private Class chunkImpl;

  /**
   * Default constructor. Used by object pool and codec.
   */
  public ChunkTypeImpl()
  {
  }

  /**
   * Creates a new chunk type.
   * 
   * @param itfSignature the signature of this chunk type
   * @param implSignature the signature of the chunk implementation
   * @throws ClassNotFoundException if the the class corresponding to the
   *           signature cannot be found.
   * @throws IllegalArgumentException if the specified signature does not
   *           correspond to an interface.
   */
  public ChunkTypeImpl(String itfSignature, String implSignature)
      throws ClassNotFoundException
  {
    this(Class.forName(itfSignature), Class.forName(implSignature));
  }

  /**
   * Creates a new Chunk type by specifying the class object of the interface
   * provided by chunks of this type
   * 
   * @param chunkItf class object of the interface provided by chunks of this
   *          type
   * @param chunkImpl class object of the implementation of the chunk.
   * @throws IllegalArgumentException if the specified class object does not
   *           designate an interface.
   */
  public ChunkTypeImpl(Class chunkItf, Class chunkImpl)
  {
    if (!chunkItf.isInterface())
    {
      throw new IllegalArgumentException(chunkItf.getName()
          + " is not an Interface");
    }
    if (!chunkItf.isAssignableFrom(chunkImpl))
    {
      throw new IllegalArgumentException(chunkImpl.getName()
          + "is not a sub class of the given interface");
    }
    this.chunkItf = chunkItf;
    this.chunkImpl = chunkImpl;
  }

  // ---------------------------------------------------------------------------
  // Implementation of the ChunkType interface.
  // ---------------------------------------------------------------------------

  /**
   * @see ChunkType#getChunkSignature()
   */
  public String getChunkSignature()
  {
    return chunkItf.getName();
  }

  /**
   * @see ChunkType#isSubTypeOf(ChunkType)
   */
  public boolean isSubTypeOf(ChunkType t)
  {
    if (t instanceof ChunkTypeImpl)
    {
      return ((ChunkTypeImpl) t).getChunkItf().isAssignableFrom(chunkItf);
    }
    else
    {
      try
      {
        Class tChunkItf = Class.forName(t.getChunkSignature());
        return tChunkItf.isAssignableFrom(chunkItf);
      }
      catch (ClassNotFoundException e)
      {
        return false;
      }
    }
  }

  // ---------------------------------------------------------------------------
  // Additional Methods
  // ---------------------------------------------------------------------------

  /**
   * Returns the Class object corresponding to the signature of this chunk type
   * 
   * @return the Class object corresponding to the signature of this chunk type
   */
  public Class getChunkItf()
  {
    return chunkItf;
  }

  /**
   * Returns the Class object implementing the chunk interface.
   * 
   * @return the Class object implementing the chunk interface.
   */
  public Class getChunkImpl()
  {
    return chunkImpl;
  }

  // ---------------------------------------------------------------------------
  // Implementation of the Recyclable interface.
  // ---------------------------------------------------------------------------

  /**
   * @see Recyclable#recycle()
   */
  public void recycle()
  {
    chunkImpl = null;
    chunkItf = null;
  }

  // ---------------------------------------------------------------------------
  // Implementation of the serialization methods
  // ---------------------------------------------------------------------------

  /**
   * @see Externalizable#readExternal(ObjectInput)
   */
  public void readExternal(ObjectInput in) throws IOException,
      ClassNotFoundException
  {
    chunkImpl = Class.forName(in.readUTF());
    chunkItf = Class.forName(in.readUTF());
  }

  /**
   * @see java.io.Externalizable#writeExternal(java.io.ObjectOutput)
   */
  public void writeExternal(ObjectOutput out) throws IOException
  {
    out.writeUTF(chunkImpl.getName());
    out.writeUTF(chunkItf.getName());
  }
}