/**
 * Dream
 * Copyright (C) 2003-2004 INRIA Rhone-Alpes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: dream@objectweb.org
 *
 * Initial developer(s): Vivien Quema
 * Contributor(s): 
 */

package org.objectweb.dream.pool;

import org.objectweb.dream.AbstractComponent;
import org.objectweb.dream.util.EmptyStringArray;
import org.objectweb.dream.util.Error;
import org.objectweb.util.monolog.api.BasicLevel;

/**
 * This class provides an implementation of message managers for non extensible
 * messages. It pools message instances.
 * <p>
 * The class of messages to be managed is given using the
 * <code>NonExtensibleMessageManagerAttributeController</code>.
 */
public class ObjectPoolImpl extends AbstractComponent
    implements
      ObjectPool,
      ObjectPoolAttributeController
{
  /** The pool of objects. */
  private Recyclable   pool[];

  /** The number of elements stored in the pool. */
  private int          elementCount = 0;

  /** The capacity of the pool. */
  private int          capacity;

  /** The name of the class of managed objects. */
  private String       className    = null;

  /** The class of objects managed by this pool. */
  private Class        objectClass;

  private final Object lock         = new Object();

  // ---------------------------------------------------------------------------
  // Implementation of the ObjectPool interface
  // ---------------------------------------------------------------------------

  /**
   * @see ObjectPool#newInstance()
   */
  public Recyclable newInstance()
  {
    Recyclable obj = null;
    synchronized (lock)
    {
      if (elementCount == 0)
      {
        if (logger.isLoggable(BasicLevel.DEBUG))
        {
          logger.log(BasicLevel.DEBUG, "Creates new instance of class "
              + className);
        }
        try
        {
          obj = (Recyclable) objectClass.newInstance();
        }
        catch (Exception e)
        {
          // Should not happen since it has been tested in the setClassName
          // method
          Error.bug(logger, e);
        }
      }
      else
      {
        elementCount -= 1;
        obj = pool[elementCount];
        pool[elementCount] = null;
      }
    }
    return obj;
  }

  /**
   * @see ObjectPool#recycleInstance(Recyclable)
   */
  public void recycleInstance(Recyclable recyclable)
  {
    if (recyclable != null && objectClass.isInstance(recyclable))
    {
      recyclable.recycle();
      synchronized (lock)
      {
        if (elementCount == pool.length)
        {
          if (logger.isLoggable(BasicLevel.WARN))
          {
            logger.log(BasicLevel.WARN, "Pool Full");
          }
          return;
        }
        pool[elementCount] = recyclable;
        elementCount += 1;
      }
    }
  }

  // ---------------------------------------------------------------------------
  // Implementation of the PoolAttributeController interface
  // ---------------------------------------------------------------------------

  /**
   * @see ObjectPoolAttributeController#getCapacity()
   */
  public int getCapacity()
  {
    return capacity;
  }

  /**
   * @see ObjectPoolAttributeController#setCapacity(int)
   */
  public void setCapacity(int capacity)
  {
    // TODO try to re-use previous instance
    synchronized (lock)
    {
      this.capacity = capacity;
      pool = new Recyclable[capacity];
      elementCount = 0;
    }
  }

  /**
   * @see ObjectPoolAttributeController#getObjectClassName()
   */
  public String getObjectClassName()
  {
    return className;
  }

  /**
   * @see ObjectPoolAttributeController#setObjectClassName(String)
   */
  public void setObjectClassName(String name)
  {
    if (className != null)
    {
      logger.log(BasicLevel.ERROR,
          "This object pool already manages objects of class " + className);
      new IllegalArgumentException(
          "This object pool already manages objects of class " + className);
    }
    this.className = name;
    try
    {
      objectClass = Class.forName(name);
      // Test that the class is correct
      Recyclable r = (Recyclable) objectClass.newInstance();
    }
    catch (Exception e)
    {
      logger.log(BasicLevel.ERROR, "Invalid class " + className, e);
      new IllegalArgumentException("Invalid class " + className);
    }
  }

  // ---------------------------------------------------------------------------
  // Implementation of the BindingController interface
  // ---------------------------------------------------------------------------
  /**
   * @see org.objectweb.fractal.api.control.BindingController#listFc()
   */
  public String[] listFc()
  {
    return EmptyStringArray.EMPTY_STRING_ARRAY;
  }

}